/* 
 * This file is part of the "OCAPI" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: John Chamberlain <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

/* used for binary content, may have nulls in the content */
typedef struct {
    char *pcharContent;		/* not necessarily a valid string--may have nulls in it */
    unsigned int iCapacity;
    unsigned int iContentLength;
    unsigned int iContentPosition0;	/* zero based */
} Buffer;

/*  used for normal strings, cannot have nulls in content*/
typedef struct {
    unsigned int iBufferSize;
    unsigned int iContentLength;
    char *sContent;
} StringBuffer;

void sbAppend(StringBuffer * sb, const char *sToBeAppended);
void sbAppend_int(StringBuffer * sb, const char *sFormatString,
		  const int i);
void sbAppend_char(StringBuffer * sb, const char c);
void sbAppend_string(StringBuffer * sb, const char *sFormatString,
		     char *s);
void sbAppend_pointer(StringBuffer * sb, const char *sFormatString,
		      void *p);
void sbInsert(StringBuffer * sb, const char *sToBeInserted);
void sbInsert_at(StringBuffer * sb, const char *sToBeInserted,
		 const unsigned int pos);
void sbInsert_int(StringBuffer * sb, const char *sFormatString,
		  const int i);
void sbInsert_string(StringBuffer * sb, const char *sFormatString,
		     const char *s);
int sbCompare(StringBuffer * sb, const char *sComparisonString);
StringBuffer *sbNew(const char *sStringOrNULL);
StringBuffer *sbNew_int(const char *sStringOrNULL, int iValue);
StringBuffer *sbNew_size(unsigned int iBufferSize);
char *sbCopyContent(StringBuffer * sb);
char *sbCopyContentSection(StringBuffer * sb, unsigned int pos0,
			   unsigned int length);
void sbFree(StringBuffer * sb);
void sbClear(StringBuffer * sb);
void sbDump(StringBuffer * sb, FILE * file);

Buffer *bufferNew(unsigned int iCapacity);
int bufferAppend(Buffer * pBuffer, char *pDataToAppend,
		 unsigned int posStart, unsigned int iLength);
void bufferFree(Buffer * pBuffer, char **sError);	/* if *sError != NULL there was an error */
void bufferDump(Buffer * pBuffer, FILE * file, unsigned int posFrom,
		unsigned int posTo);
char *bufferCopyAsString(Buffer * pBuffer);
int bufferStartsWith(Buffer * bufferText, char *sSearchString);
Buffer *bufferLoadFile(const char *sFilename);

void vTestInsert(void);
