/* 
 * This file is part of the "Ocapi" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: Dennis Heimbigner
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

// -*- mode: c++; c-basic-offset:4 -*-

// This file is part of libdap, A C++ implementation of the OPeNDAP Data
// Access Protocol.

// Copyright (c) 2002,2003 OPeNDAP, Inc.
// Author: James Gallagher <jgallagher@opendap.org>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 
// (c) COPYRIGHT URI/MIT 1994-1999
// Please read the full copyright statement in the file COPYRIGHT_URI.
//
// Authors:
//      jhrg,jimg       James Gallagher <jgallagher@gso.uri.edu>

/*
   Grammar for the DDS. This grammar can be used with the bison parser
   generator to build a parser for the DDS. It assumes that a scanner called
   `ddslex()' exists and returns several token types (see das.tab.h)
   in addition to several single character token types. The matched lexeme
   for an ID is stored by the scanner in a global char * `ddslval'.
   Because the scanner returns a value via this global and because the parser
   stores ddslval (not the information pointed to), the values of rule
   components must be stored as they are parsed and used once accumulated at
   or near the end of a rule. If ddslval returned a value (instead of a
   pointer to a value) this would not be necessary.

   jhrg 8/29/94 
*/

%{
#include "parse.h"

#define YYERROR_VERBOSE 1
//#define YYPARSE_PARAM arg

int dapsetup(void);

ParserBuffer* dapParserState;

Stack* dapdimnames;
Stack* dapdimsizes;
Stack* dapstack;
Stack* daptemp;
Stack* dapvalues;

%}


%union {
    char word[ID_MAX];
    eNODE_TYPE nodetype;
    int counter;
}

%token <word> SCAN_WORD
%token <word> SCAN_BYTE
%token <word> SCAN_INT16
%token <word> SCAN_UINT16
%token <word> SCAN_INT32
%token <word> SCAN_UINT32
%token <word> SCAN_FLOAT32
%token <word> SCAN_FLOAT64
%token <word> SCAN_STRING
%token <word> SCAN_URL 
%token <word> SCAN_DATASET
%token <word> SCAN_SEQUENCE
%token <word> SCAN_STRUCTURE
%token <word> SCAN_FUNCTION
%token <word> SCAN_GRID
%token <word> SCAN_ARRAY
%token <word> SCAN_MAPS 
%token <word> SCAN_ATTR
%token <word> SCAN_ALIAS 
%token <word> SCAN_DATABEGIN

%type <word> var datasetname var_name name
%type <word> url str_or_id float_or_int
%type <nodetype> base_type
%type <counter> declarations attr_list
%%

start:
	  SCAN_DATASET {dapsetup();}
            datasetbody
	| SCAN_ATTR {dapsetup();}
	  attributebody
	;
	
datasetbody:
	  '{' declarations '}' datasetname ';'
	    {
		makeRoot(STRUCTURE_DDS,$4,$2);	
	    }
        | error
                {
		    dap_parse_error("Illegal dataset body", "dataset",dap_line_num);
		    dap_cleanup();
		    YYABORT;
		}
	;

declarations:
	  /* empty */ {$$ = 0;}
        | declarations declaration {$$=$1+1;}
	;

declaration:
	  base_type var ';' {makeArray($2,$1);}
	| SCAN_STRUCTURE '{' declarations '}' var ';'
	    {makeStructure($5,$3);}
	| SCAN_SEQUENCE '{' declarations '}' var ';'
	    {makeSequence($5,$3);}
	| SCAN_GRID
	  '{' SCAN_ARRAY ':' declaration SCAN_MAPS ':' declarations '}' var ';'
	    {makeGrid($10,$8);}
        | error 
                {
		    dap_parse_error("Unrecognized declaration kind", NULL, dap_line_num);
		    YYABORT;
		}
	;
 

base_type:
	  SCAN_BYTE {$$ = NODE_Byte;}
	| SCAN_INT16 {$$ = NODE_Int16;}
	| SCAN_UINT16 {$$ = NODE_UInt16;}
	| SCAN_INT32 {$$ = NODE_Int32;}
	| SCAN_UINT32 {$$ = NODE_UInt32;}
	| SCAN_FLOAT32 {$$ = NODE_Float32;}
	| SCAN_FLOAT64 {$$ = NODE_Float64;}
	| SCAN_URL {$$ = NODE_URL;}
	| SCAN_STRING {$$ = NODE_String;}
	;

var:
 	var_name array_decls {strcpy($$,$1);}
	;

array_decls:
	  /* empty */
	     {
		stackClear(dapdimnames); stackClear(dapdimsizes);
	     }
	| array_decls array_decl
	;

array_decl:
	   '[' SCAN_WORD ']'
	    {
		if(!check_int32($2))
		    dap_parse_error("Dimension not an integer", $2, dap_line_num);
		stackPush(dapdimnames,(elem_t)NULL);
		stackPush(dapdimsizes,(elem_t)atol($2));
	    }
	 | '[' SCAN_WORD '=' SCAN_WORD ']'
	    {
		if(!check_int32($4))
			dap_parse_error("Dimension not an integer",$4, dap_line_num);
		stackPush(dapdimnames,(elem_t)nulldup($2));
		stackPush(dapdimsizes,(elem_t)atol($4));
	    }
	 | error
                 {
		     dap_parse_error("Illegal dimension declaration", NULL, dap_line_num);
		     YYABORT;
		 }
	;

datasetname:
	  var_name {strcpy($$,$1);}
	| SCAN_DATASET {strcpy($$,$1);}
        | error 
                {
		     dap_parse_error("Illegal dataset name", NULL, dap_line_num);
		     YYABORT;
		}
	;

var_name: name {strcpy($$,$1);}

attributebody:
	  '{' attr_list '}'
	        {
		makeAttrroot($2);
		}
	| error
                {
		    dap_parse_error("Malformed DAS body", NULL, dap_line_num);
		}
	;

attr_list:
	  /* empty */ {$$=stackDepth(dapstack);}  /* store the stackbase */
	| attr_list attr_tuple
	    {
		if(checkdupattr($1)) {
		    dap_parse_error("Duplicate attribute names at same level (1)",
					NULL,dap_line_num);
		}
		$$=$1;
	    }
	;

attr_tuple:
	  alias
        | SCAN_BYTE name bytes ';'
	    {makeAttr($2,ATTRIBUTE_Byte);}
	| SCAN_INT16 name int16 ';'
	    {makeAttr($2,ATTRIBUTE_Int16);}
	| SCAN_UINT16 name uint16 ';'
	    {makeAttr($2,ATTRIBUTE_UInt16);}
	| SCAN_INT32 name int32 ';'
	    {makeAttr($2,ATTRIBUTE_Int32);}
	| SCAN_UINT32 name uint32 ';'
	    {makeAttr($2,ATTRIBUTE_UInt32);}
	| SCAN_FLOAT32 name float32 ';'
	    {makeAttr($2,ATTRIBUTE_Float32);}
	| SCAN_FLOAT64 name float64 ';'
	    {makeAttr($2,ATTRIBUTE_Float64);}
	| SCAN_STRING name strs ';'
	    {makeAttr($2,ATTRIBUTE_String);}
	| SCAN_URL name urls ';'
	    {makeAttr($2,ATTRIBUTE_URL);}
	| SCAN_WORD '{' attr_list '}'
	    {
		makeAttrnode($1,$3);
	    }
	| error 
            { 
	        dap_parse_error("Malformed attribute",  NULL, dap_line_num);
	    }
	;

bytes:
	  SCAN_WORD
	    {stackClear(dapvalues); add_attributevalue(nulldup($1),NODE_Byte);}
	| bytes ',' SCAN_WORD 
	    {add_attributevalue(nulldup($3),NODE_Byte);}
	;

int16:
	  SCAN_WORD
	    {stackClear(dapvalues); add_attributevalue(nulldup($1),NODE_Int16);}
	| int16 ',' SCAN_WORD
	    {add_attributevalue(nulldup($3),NODE_Int16);}
	;

uint16:
	  SCAN_WORD
	    {stackClear(dapvalues); add_attributevalue(nulldup($1),NODE_UInt16);}
	| uint16 ',' SCAN_WORD
	    {add_attributevalue(nulldup($3),NODE_UInt16);}
	;

int32:
	  SCAN_WORD
	    {stackClear(dapvalues); add_attributevalue(nulldup($1),NODE_Int32);}
	| int32 ',' SCAN_WORD
	    {add_attributevalue(nulldup($3),NODE_Int32);}
	;

uint32:
	  SCAN_WORD
	    {stackClear(dapvalues); add_attributevalue(nulldup($1),NODE_UInt32);}
	| uint32 ',' SCAN_WORD
	    {add_attributevalue(nulldup($3),NODE_UInt32);}
	;

float32:
	  float_or_int
	    {stackClear(dapvalues); add_attributevalue(nulldup($1),NODE_Float32);}
	| float32 ',' float_or_int
	    {add_attributevalue(nulldup($3),NODE_Float32);}
	;

float64:
	  float_or_int
	    {stackClear(dapvalues); add_attributevalue(nulldup($1),NODE_Float64);}
	| float64 ',' float_or_int
	    {add_attributevalue(nulldup($3),NODE_Float64);}
	;

strs:
	  str_or_id
	    {stackClear(dapvalues); add_attributevalue(nulldup($1),NODE_String);}
	| strs ',' str_or_id
	    {add_attributevalue(nulldup($3),NODE_String);}
	;

urls:
	  url
	    {stackClear(dapvalues); add_attributevalue(nulldup($1),NODE_URL);}
	| urls ',' url
	    {add_attributevalue(nulldup($3),NODE_URL);}
	;

url:
	SCAN_WORD {strcpy($$,$1);}
	;

str_or_id:
	SCAN_WORD {strcpy($$,$1);}
	;

float_or_int:
	SCAN_WORD {strcpy($$,$1);}
	;

alias:
	SCAN_ALIAS SCAN_WORD SCAN_WORD
	    {
		dap_parse_error("Alias not currently supported",$2, dap_line_num);
		YYABORT;
	    }
	;

/* Note that variable names like "byte" are legal names
   and are disambiguated by context
*/
name:
          SCAN_WORD {strcpy($$,$1);}
	| SCAN_BYTE {strcpy($$,$1);}
	| SCAN_INT16 {strcpy($$,$1);}
	| SCAN_INT32 {strcpy($$,$1);}
	| SCAN_UINT16 {strcpy($$,$1);}
	| SCAN_UINT32 {strcpy($$,$1);}
	| SCAN_FLOAT32 {strcpy($$,$1);}
	| SCAN_FLOAT64 {strcpy($$,$1);}
	| SCAN_STRING {strcpy($$,$1);}
	| SCAN_URL {strcpy($$,$1);}
	| SCAN_STRUCTURE {strcpy($$,$1);}
	| SCAN_SEQUENCE {strcpy($$,$1);}
	| SCAN_GRID {strcpy($$,$1);}
	| SCAN_ARRAY {strcpy($$,$1);}
	| SCAN_MAPS {strcpy($$,$1);}
	| SCAN_ATTR {strcpy($$,$1);}
	;

%%

int yyerror(char* msg)
{
    dap_parse_error(msg,NULL, dap_line_num);
    return 0;
}

int
dapsetup(void)
{
    if(dapdimnames == NULL) {
	dapdimnames = stackNew(STACKALLOC);
	dapdimsizes = stackNew(STACKALLOC);
        daptemp = stackNew(STACKALLOC);
        dapstack = stackNew(STACKALLOC);
	dapvalues = stackNew(STACKALLOC);
    } else {
	stackClear(dapdimnames);
	stackClear(dapdimsizes);
	stackClear(daptemp);
	stackClear(dapstack);
	stackClear(dapvalues);
    }
    return 1;
}
    

int check_int32(char* val)
{
    char* ptr;
    int ok = 1;
    long iv = strtol(val,&ptr,0); // 0=>auto determine base
    if((iv == 0 && val == ptr) || *ptr != '\0') ok=0;
    else if(iv > DODS_INT_MAX || iv < DODS_INT_MIN) ok=0;
    return ok;
}
