
// Copyright (c) 2008 OPeNDAP, Inc.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.

// #include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifndef WIN32
#include <unistd.h>
#else
#include "getopt.h"
#endif

#include "ocapi.h"


void test_print_das(ThreadSpecificStorage* TSS, char *sURL); /*Connect to the test server and retrieve das metadata*/
void test_print_dds(ThreadSpecificStorage* TSS, char *sURL, char *constraint); /*Connect to the test server and retrieve dds metadata*/
void test_print_data(ThreadSpecificStorage* TSS, char *sURL, char *constraint); /*Connect to the test server and retrieve data from server*/


int main (int argc, char *argv[]){
      char* sError = NULL;
      ThreadSpecificStorage *TSS = NULL;
      int c;
      int get_das = 0, get_dds = 0, get_data = 0;
      char *constraint = 0;
      int constraint_len = 0;
      char *sURL = 0;
      int i;
      int len;
      int status = EXIT_FAILURE;

      int iConnectTimeout_seconds = -1;
      int iReadTimeout_seconds = -1;
      char *sProtocol = NULL;
      FILE *pLoggingStream = NULL;
      char *sUserAgent = NULL;

      char *sHost = NULL;
      int iPort = -1;
      char *sUser = NULL;
      char *sPassword = NULL;
      char *sUserC = NULL;
      char *sPasswordC = NULL;
      int sVerify = 0;

      for (i = 0; i < argc; ++i)

      /*Retrieving option then move to specified option in code*/
        while ((c = getopt(argc, argv, ":adDc:hp:P:n:w:m:r:s")) != -1) {
	    switch (c) {
	    case 'a':
		get_das = 1;
		break;
	    case 'd':
		get_dds = 1;
		break;
	    case 'D':
		get_data = 1;
		break;
	    case 'c':
		constraint_len = strlen(optarg);
		constraint = malloc((constraint_len + 1) * sizeof(char));
		if (!constraint) {
#if 0
		    fprintf(stderr, "Could not allocate memory (%c).\n", optopt);
#endif
		    status = EXIT_FAILURE;
		    goto exit;
		} else {
#if 0
		    fprintf(stderr, "Constraint: %s\n", optarg);
#endif
		    strncpy(constraint, optarg, constraint_len + 1);
		    constraint[constraint_len] = '\0';
		}
		break;
	    case 'p': {
		int sHost_len = strlen(optarg);
		sHost = malloc((sHost_len + 1) * sizeof(char));
		if (!sHost) {
#if 0
		    fprintf(stderr, "Could not allocate memory (%c).\n", optopt);
#endif
		    status = EXIT_FAILURE;
		    goto exit;
		} else {
		    strncpy(sHost, optarg, sHost_len + 1);
		    sHost[sHost_len] = '\0';
#if 0
		    fprintf(stderr, "Proxy host: %s\n", sHost);
#endif
		}
		break;
	    }
	    case 'P':
		iPort = atoi(optarg);
#if 0
		fprintf(stderr, "Proxy port: %d\n", iPort);
#endif
		break;
	    case 'n': {
		int sUser_len = strlen(optarg);
		sUser = malloc((sUser_len + 1) * sizeof(char));
		if (!sUser) {
#if 0
		    fprintf(stderr, "Could not allocate memory (%c).\n", optopt);
#endif
		    status = EXIT_FAILURE;
		    goto exit;
		} else {
		    strncpy(sUser, optarg, sUser_len + 1);
		    sUser[sUser_len] = '\0';
#if 0
		    fprintf(stderr, "Username: %s\n", sUser);
#endif
		}
		break;
	    }
	    case 'w': {
		int sPassword_len = strlen(optarg);
		sPassword = malloc((sPassword_len + 1) * sizeof(char));
		if (!sPassword) {
#if 0
		    fprintf(stderr, "Could not allocate memory (%c).\n", optopt);
#endif
		    status = EXIT_FAILURE;
		    goto exit;
		} else {
		    strncpy(sPassword, optarg, sPassword_len + 1);
		    sPassword[sPassword_len] = '\0';
#if 0
		    fprintf(stderr, "Password: %s\n", sPassword);
#endif
		}
		break;
	    }
	    case 's': {
		sVerify = 1;
		break;
	    }
	    case 'r': {
		int sUserC_len = strlen(optarg);
		sUserC = malloc((sUserC_len + 1) * sizeof(char));
		if (!sUserC) {
#if 0
		    fprintf(stderr, "Could not allocate memory (%c).\n", optopt);
#endif
		    status = EXIT_FAILURE;
		    goto exit;
		} else {
		    strncpy(sUserC, optarg, sUserC_len + 1);
		    sUserC[sUserC_len] = '\0';
#if 0
		    fprintf(stderr, "Username: %s\n", sUserC);
#endif
		}
		break;
	    }
	    case 'm': {
		int sPasswordC_len = strlen(optarg);
		sPasswordC = malloc((sPasswordC_len + 1) * sizeof(char));
		if (!sPasswordC) {
#if 0
		    fprintf(stderr, "Could not allocate memory (%c).\n", optopt);
#endif
		    status = EXIT_FAILURE;
		    goto exit;
		} else {
		    strncpy(sPasswordC, optarg, sPasswordC_len + 1);
		    sPasswordC[sPasswordC_len] = '\0';
#if 0
		    fprintf(stderr, "Password: %s\n", sPasswordC);
#endif
		}
		break;
	    }
	    case 'h':
#if 0
	        fprintf(stderr, "Help... (%c)\n", optopt);
#endif
		status = EXIT_SUCCESS;
		goto exit;
		break;
	    case '?':
#if 0
		fprintf(stderr, "unrecognized option: -%c\n", optopt);
#endif
		status = EXIT_SUCCESS;
		goto exit;
		break;
	    case ':':
#if 0
		fprintf(stderr, "missing argument: -%c\n", optopt);
#endif
		status = EXIT_FAILURE;
		goto exit;
		break;
	    default:
#if 0
		fprintf(stderr, "bad option");
#endif
		status = EXIT_FAILURE;
		goto exit;
	    }
	}

      /*If no option given print out message*/
      if (!get_das && !get_dds && !get_data) {
#if 0
    	  fprintf(stderr, "No option given.\n");
#endif
    	  status = EXIT_FAILURE;
    	  goto exit;
      }

      /* Initialize the Ocapi Library: This allocate the thread-specific
       * storage. */

      if (zInitializeOCAPI(&TSS, &sError) == FAILURE) {
#if 0
          fprintf(stderr, "Failed to initialize Ocapi %s\n", sError);
#endif
  	  status = EXIT_FAILURE;
  	  goto exit;
      }

      if (configure(TSS, iConnectTimeout_seconds, iReadTimeout_seconds, sProtocol, pLoggingStream,
          sUserAgent, &sError) == FAILURE){
#if 0
    	  fprintf(stderr, "Failed to configure the connection parameters %s\n",sError);
#endif
      status = EXIT_FAILURE;
      goto exit;
      }

      if (configure_proxy(TSS, sHost, iPort, sUser, sPassword, &sError) == FAILURE){
#if 0
    	  fprintf(stderr, "Failed to configure the connection parameters for proxy server %s\n", sError);
#endif
      status = EXIT_FAILURE;
      goto exit;
      }

      if(configure_httpAuth(TSS, sUserC, sPasswordC, &sError) == FAILURE){
#if 0
    	  fprintf(stderr, "Failed to configure the connection parameters for HTTP %s\n", sError);
#endif
      status = EXIT_FAILURE;
      goto exit;
      }

      if (sVerify){
    	  if(configure_ssl(TSS, sVerify, &sError) == FAILURE) {
#if 0
          fprintf(stderr, "Failed to shut off ssl configuration %s\n", sError);
#endif
          status = EXIT_FAILURE;
          goto exit;
    	  }
      }

      for (i = optind; i < argc; ++i) {

	/*Allocates additional memory for specified structure type and URL including ending null value: \0*/
	len = strlen(argv[i]);
	sURL = malloc((len + 1) * sizeof(char));
	if (!sURL) {
#if 0
	    fprintf(stderr, "Could not allocate memory.");
#endif
	    status = EXIT_FAILURE;
	    goto exit;
	}
	strncpy(sURL, argv[i], len + 1);
	sURL[len] = '\0';

	/*Move to the specified function from the option chosen above*/
	if (get_das)
	    test_print_das(TSS, argv[i]);

	if (get_dds)
	    test_print_dds(TSS, sURL, constraint);

	if (get_data)
	    test_print_data(TSS, sURL, constraint);

	if (sURL)
	    free(sURL);
    }

    status = EXIT_SUCCESS;

  exit:

    if (constraint)
	free(constraint);
    if (sHost)
	free(sHost);
    if (sUser)
	free(sUser);
    if (sPassword)
	free(sPassword);
    if (sUserC)
	free(sUserC);
    if (sPasswordC)
	free(sPasswordC);
    vCleanupOCAPI(TSS);

    return status;
}

/*Locates a specified Url and retrieves data from that URL*/
void test_print_data(ThreadSpecificStorage* TSS, char *sURL, char *constraint) {
    char* sError = NULL;
    URL* pUrl = NULL;
    DAPObject* Data = NULL;
    eRESULT eResult;
    int len;
    char *sFullURL = NULL;

    /*Allocates memory for the specified Url.*/
    pUrl = malloc(sizeof(URL));
    if (!pUrl) {
#if 0
	fprintf(stderr, "Could not allocate memory.\n");
#endif
	goto exit;
    }

    /*If a constraint is provided then add constraint*/
    if (constraint != NULL) {
	/*Allocate additional memory for specified structure type including ending null value: \0*/
	len = strlen(sURL) + strlen(constraint);
	sFullURL = malloc((len + 5 + 1 + 1) * sizeof(char));
	if (!sFullURL) {
#if 0
	    fprintf(stderr, "Could not allocate memory.\n");
#endif
	    goto exit;
	} else {
	    /*Copies the given URL into the sFullURL and adds memory for the ending null value
	     *otherwise the memory length will not be null terminated*/
	    strncpy(sFullURL, sURL, len + 1);
	    sFullURL[len] = '\0';
	    /*Concatinates the URL with the specified data type, does not need terminating null value*/
	    strncat(sFullURL, ".dods", 5);
	    strncat(sFullURL, "?", 1);
	    strncat(sFullURL, constraint, strlen(constraint));
	    /*Ocapi's official form of the URL to set the given URL (sFullURL) to pURL*/

	}
    } else { /*If there is no constraint provided*/
	len = strlen(sURL);
	sFullURL = malloc((len + 5 + 1) * sizeof(char));
	if (!sFullURL) {
#if 0
	    fprintf(stderr, "Could not allocate memory.\n");
#endif
	    goto exit;
	} else {
	    strncpy(sFullURL, sURL, len + 1);
	    sFullURL[len] = '\0';
	    strncat(sFullURL, ".dods", 5);

	}
    }
    /*Ocapi's official form of the URL to set the given URL (sFullURL) to pURL*/
    pUrl->sCanonicalURL = sFullURL;

    /*Retrieves the DATA from the above Url.*/
    eResult = structure_Load_URL(pUrl, &Data, TSS, &sError);
    if (eResult == FAILURE) {
	fputs("The data did not load. ", stderr);
	if (sError != NULL)
	    fputs(sError, stderr);
	else
	    fputs("No Error message given.", stderr);
	fputs("\n", stderr);
	goto exit;
    }

    /*Print the DATA to a file*/
    eResult = structure_Print_ASCII_toFile(Data, stdout, &sError);
    if (eResult == FAILURE) {
	fputs("The data did not print. ", stderr);
	if (sError != NULL)
	    fputs(sError, stderr);
	else
	    fputs("No error message given.", stderr);
	fputs("\n", stderr);
	goto exit;
    }
    /*refer to UserGuide.txt line 217 and look at example_main.c -- the following is not working, structure not being freed*/
    /* free memory when done with structure */
    {
	int status = structure_Free(Data, &sError);
	if ( /*structure_Free( Data, &sError )*/status != SUCCESS) {
	    fputs("error freeing data: ", stderr );
	    fputs(sError, stderr );
	    fputs("\n", stderr );
	}
    }
    goto exit;

    exit:
    /*Free up memory */
    if (sError != NULL)
	free(sError);
    if (pUrl != NULL)
	free(pUrl);
    if (sFullURL != NULL)
	free(sFullURL);

}
/*Locates a specified Url and retrieves DAS metadata from that URL*/
void test_print_das(ThreadSpecificStorage* TSS, char *sURL) {
    char* sError = NULL;
    URL* pUrl = NULL;
    DAPObject* Data = NULL;
    eRESULT eResult;
    int len;
    char *sFullURL = NULL;

    /*Allocates memory for the specified Url.*/
    pUrl = malloc(sizeof(URL));
    if (!pUrl) {
#if 0
	fprintf(stderr, "Could not allocate memory.");
#endif
	goto exit;
    }

    /*Allocates additional memory for specified structure type including ending null value: \0*/
    len = strlen(sURL);
    sFullURL = malloc((len + 4 + 1) * sizeof(char));
    if (!sFullURL) {
#if 0
	fprintf(stderr, "Could not allocate memory.");
#endif
	goto exit;
    } else {
	strncpy(sFullURL, sURL, len + 1);
	sFullURL[len] = '\0';
	strncat(sFullURL, ".das", 4);

    }

    /*Ocapi's official form of the URL to set the given URL (sFullURL) to pURL*/
    pUrl->sCanonicalURL = sFullURL;

    /*Retrieves the DATA from the above Url.*/
    eResult = structure_Load_URL(pUrl, &Data, TSS, &sError);
    if (eResult == FAILURE) {
	fputs("The data did not load.", stderr);
	if (sError != NULL)
	    fputs(sError, stderr);
	else
	    fputs("No error message given.", stderr);
	goto exit;
    }
    /*Print the DATA to a file*/
    eResult = structure_Print_ASCII_toFile(Data, stdout, &sError);
    if (eResult == FAILURE) {
	fputs("The data did not print.", stderr);
	if (sError != NULL)
	    fputs(sError, stderr);
	else
	    fputs("No error message given.", stderr);
	goto exit;
    }
    /*refer to UserGuide.txt line 217 and look at example_main.c -- the following is not working, structure not being freed*/
    /* free memory when done with structure */
    {
	int status = structure_Free(Data, &sError);
	if ( /*structure_Free( Data, &sError )*/status != SUCCESS) {
	    fputs("error freeing data: ", stderr );
	    fputs(sError, stderr );
	    fputs("\n", stderr );
	}
    }
    goto exit;

    exit:
    /*Free up memory */
    if (sError != NULL)
	free(sError);
    if (pUrl != NULL)
	free(pUrl);
    if (sFullURL != NULL)
	free(sFullURL);

}

/*Locates a specified Url and retrieves DDS metadata from that URL*/
void test_print_dds(ThreadSpecificStorage* TSS, char *sURL, char *constraint) {
    char* sError = NULL;
    URL* pUrl = NULL;
    DAPObject* Data = NULL;
    eRESULT eResult;
    int len;
    char *sFullURL = NULL;

    /*Allocates memory for the specified Url.*/
    pUrl = malloc(sizeof(URL));
    if (!pUrl) {
#if 0
	fprintf(stderr, "Could not allocate memory.");
#endif
	goto exit;
    }

    /*If a constraint is provided then add constraint*/
    if (constraint != NULL) {
	/*Allocates additional memory for specified structure type including ending null value: \0*/
	len = strlen(sURL) + strlen(constraint);
	sFullURL = malloc((len + 4 + 1 + 1) * sizeof(char));
	if (!sFullURL) {
#if 0
	    fprintf(stderr, "Could not allocate memory.");
#endif
	    goto exit;
	} else {
	    /*Copies the given URL into the sFullURL and adds memory for the ending null value
	     *otherwise the memory length will not be null terminated*/
	    strncpy(sFullURL, sURL, len + 1);
	    sFullURL[len] = '\0';
	    /*Concatinates the URL with the specified data type, does not need terminating null value*/
	    strncat(sFullURL, ".dds", 4);
	    strncat(sFullURL, "?", 1);
	    strncat(sFullURL, constraint, strlen(constraint));

	}
    } else { /*If there is no constraint provided*/
	len = strlen(sURL);
	sFullURL = malloc((len + 4 + 1) * sizeof(char));
	if (!sFullURL) {
#if 0
	    fprintf(stderr, "Could not allocate memory.");
#endif
	    goto exit;
	} else {
	    strncpy(sFullURL, sURL, len + 1);
	    sFullURL[len] = '\0';
	    strncat(sFullURL, ".dds", 4);

	}
    }

    /*Ocapi's official form of the URL to set the given URL (sURL) to pURL*/
    pUrl->sCanonicalURL = sFullURL;

    /*Retrieves the DATA from the above Url.*/
    eResult = structure_Load_URL(pUrl, &Data, TSS, &sError);
    if (eResult == FAILURE) {
	fputs("The data did not load.", stderr);
	if (sError != NULL)
	    fputs(sError, stderr);
	else
	    fputs("No error message given.", stderr);
	goto exit;
    }

    /*Print the DATA to a file*/
    eResult = structure_Print_ASCII_toFile(Data, stdout, &sError);
    if (eResult == FAILURE) {
	fputs("The data did not print.", stderr);
	if (sError != NULL)
	    fputs(sError, stderr);
	else
	    fputs("No error message given.", stderr);
	goto exit;
    }
    /*refer to UserGuide.txt line 217 and look at example_main.c -- the following is not working, structure not being freed*/
    /* free memory when done with structure */
    {
	int status = structure_Free(Data, &sError);
	if ( /*structure_Free( Data, &sError )*/status != SUCCESS) {
	    fputs("error freeing data: ", stderr );
	    fputs(sError, stderr );
	    fputs("\n", stderr );
	}
    }
    goto exit;

    exit:
    /*Free up memory */
    if (sError != NULL)
	free(sError);
    if (pUrl != NULL)
	free(pUrl);
    if (sFullURL != NULL)
	free(sFullURL);

}
