/*
 * This file is part of the "OCAPI" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: John Chamberlain <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

/* Defining HAVE_CONFIG_H in the Makefile breaks the build of http.c 12/04/08 */
/* #ifdef HAVE_CONFIG_H */
#include "config.h"
/* #endif */

#include "debug.h"
#include "ocapi.h"
#include "buffers.h"
#include "utility.h"
/*#include "stack.h"  added this to parse.h*/
#include <curl/curl.h>

#define DEBUG FALSE

#define	DODS_CHAR_BIT	8
#define	DODS_SCHAR_MIN	-128
#define	DODS_SCHAR_MAX	127
#define	DODS_UCHAR_MAX	255U
#define	DODS_UCHAR_MIN	0

#define	DODS_SHRT_MIN	-32768
#define	DODS_SHRT_MAX	32767
#define	DODS_USHRT_MAX	65535U

#define	DODS_INT_MIN	(-2147483647 - 1)
#define	DODS_INT_MAX	2147483647
#define	DODS_UINT_MAX	4294967295U

#define	DODS_LONG_MIN	(-2147483647 - 1)
#define	DODS_LONG_MAX	2147483647
#define	DODS_ULONG_MAX	4294967295UL

#define	DODS_DBL_DIG	15	/* digits of precision of a "double" */
#define	DODS_DBL_MAX	1.7976931348623157E+308	/* max decimal value of a */
						/* "double" */
#define	DODS_DBL_MIN	2.2250738585072014E-308	/* min decimal value of a */
						/* "double" */
#define	DODS_FLT_DIG	6	/* digits of precision of a "float" */
#define	DODS_FLT_MAX	3.402823466E+38F	/* max decimal value of a "float" */
#define	DODS_FLT_MIN	1.175494351E-38F	/* min decimal value of a "float" */

#define	DODS_MB_LEN_MAX	4

#define MASK_Node 0x20DE000C	/* allows easy identification of valid type codes */
#define MASK_Attribute 0x20DE000C	/* allows easy identification of valid attribute codes */

eRESULT zInitialize_http(ConnectionParameters **
			 hDefaultConnectionParameters,
			 StringBuffer * sbError);
eRESULT imp_structure_New(DAPObject ** structure, StringBuffer * sbError);
eRESULT imp_structure_Free(DAPObject * structure, StringBuffer * sbError);
eRESULT imp_node_Free(DAPNode * node, StringBuffer * sbError);
eRESULT imp_structure_Verify(DAPObject * pStructure, int zRequireParent,
			     StringBuffer * sbError);
eRESULT imp_structure_VerifyNode(DAPNode * pNode, int zRequireParent,
				 StringBuffer * sbError);
eRESULT imp_structure_VerifyArray(NodeContent_Array * pContent_array,
				  StringBuffer * sbError);
eRESULT imp_structure_VerifyGrid(NodeContent_Grid * pContent_grid,
				 StringBuffer * sbError);
eRESULT imp_structure_VerifySequence(DAPNode * pSequenceNode,
				     NodeContent_Sequence *
				     pNodeContent_sequence,
				     int zRequireParent,
				     StringBuffer * sbError);
eRESULT imp_structure_VerifyStructure(DAPNode * nodeStructure,
				      NodeContent_Structure *
				      pNodeContent_structure,
				      int zRequireParent,
				      StringBuffer * sbError);
eRESULT imp_structure_Print_ASCII_toFile(DAPObject * structure,
					 FILE * stream,
					 StringBuffer * sbError);
eRESULT imp_buffer_Free(Buffer *
			pBuffer
			/* not used jhrg 4/3/07 , StringBuffer *sbError */
			);

/* parser stuff */
#if 0
typedef enum { GM_undefined, GM_Values, GM_Maps } eGRID_MODE;
#endif
/* In buffered mode the http connection reads all the data until it is
 * complete or the mode changes. This mode is used to load short downloads
 * like DDSs. The entire buffer must be loaded before anything else is done.
 *
 * Data mode is a special mode for reading DataDDSs. In this mode the callback
 * routine used by the HTTP library (which is currently curl_callback_data() in
 * http.c) activates the data deserializer directly when it receives data.
 * In this mode the buffer only stores the chunk of data being actively parsed.
 */

typedef enum { RM_undefined, RM_Buffered, RM_Data } eREAD_MODE;

typedef struct {
    int iStatus;		/* TRUE or FALSE */
#if 0
    eGRID_MODE eGridMode;
#endif
    eREAD_MODE eReadMode;	/* reading from a buffer or stream, push, pull, etc */
    DAPObject *pOStructure;	/* points to the OPeNDAP structure being built */
    char *sError;		/* pointer to an error message */
    Buffer *pInputBuffer;
    eSTRUCTURE_TYPE eStructureType;	/* structure type as determined from the content description in the server header */
    CURL *pCurlRequest;		/* this is needed by the callback function to determine the content type */
} ParserBuffer;

int ddsparse(void *pParserState);
