/*
 * This file is part of the "OCAPI" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005-7 OPeNDAP, Inc.
 * Author: John Chamberlain <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */
/*Turn this on so the DBG statements will print for debugging purposes--remove if do not want the printout*/
/*#define DODS_DEBUG*/

/*#include "config.h"*/

#include "global.h"

#include <stdlib.h>
#include <pthread.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#define DATA_HEADER "\nData:\n"

pthread_mutex_t mutexInitializer = PTHREAD_MUTEX_INITIALIZER;
int gzDebug_HTTP = FALSE;

/* extern ConversionFormats gConversionFormats;  not used -- was defined in ocapi.c */

extern void (*gpvUpdateProgress)(int iBytesReceived, int iBytesTotal,
		char *sStatus); /* defined in main.c */
extern eRESULT read_Node(DAPNode * nodeTemplate, void *nodeContent_Supplied,
		Buffer * pBuffer, StringBuffer * sbError); /* defined in data.c */
extern eRESULT zParseDAS(DAPObject ** hStructure, Buffer * pBuffer,
		StringBuffer * sbError); /* defined in parseutil.c */
extern void vResetScanner(); /* defined in dds.lex */

eRESULT zInitialize_http(ConnectionParameters **
		hDefaultConnectionParameters, StringBuffer * sbError);
eRESULT zConnection_Fetch(char *sURL, Buffer * pBuffer,
		ThreadSpecificStorage * pTSS, StringBuffer * sbError);
eRESULT zConnection_Load(char *sURL, DAPObject ** hStructure,
		ThreadSpecificStorage * pTSS, StringBuffer * sbError);
eRESULT zConnection_FetchWithParameters(ConnectionParameters * parameters,
		Buffer * pBuffer, StringBuffer * sbError);
eRESULT zFetch_Synchronous(ConnectionParameters * parameters,
		ParserBuffer * pParserBuffer, StringBuffer * sbError);
eRESULT zFetch_Asynchronous(ConnectionParameters * parameters,
		ParserBuffer * pParserBuffer, StringBuffer * sbError);
size_t curl_callback_data(void *buffer, size_t size, size_t nmemb, void *userp);
extern char *findError(char *sBuffer); /* defined in utility.c */
char *sCURLError(int iErrorCode);
int parse_buffer(void *pBuffer);

extern unsigned char *gByte;

/* todo segregate the OPeNDAP functionality from the HTTP functionality
 * currently the easiest thing to do is to integrate the two because the HTTP
 * features are not really generic enough to stand on their own
 */
eRESULT zInitialize_http(ConnectionParameters ** hDefaultConnectionParameters,
		StringBuffer * sbError) {
	StringBuffer *sbUserAgent;
	char *sDefaultUserAgent;
	ConnectionParameters *pDefaultConnectionParameters;

	/* determine the user agent */
	sbUserAgent = sbNew("OCAPI/");
	sbAppend(sbUserAgent, getVersion());
	sDefaultUserAgent = sbCopyContent(sbUserAgent);
	sbFree(sbUserAgent);

	/* setup the default connection parameters */
	pDefaultConnectionParameters = malloc(sizeof(ConnectionParameters));
	if (pDefaultConnectionParameters == NULL) {
		sbAppend(sbError,
				"failed to allocated memory for default connection parameters");
		*hDefaultConnectionParameters = NULL;
		return FAILURE;
	}

	pDefaultConnectionParameters->pstructProxy = NULL;
	pDefaultConnectionParameters->sUserC = NULL;
	pDefaultConnectionParameters->sPasswordC = NULL;
	pDefaultConnectionParameters->iConnectTimeout_seconds
			= DEFAULT_TimeoutConnect_seconds;
	pDefaultConnectionParameters->iReadTimeout_seconds
			= DEFAULT_TimeoutRead_seconds;
	pDefaultConnectionParameters->pLoggingStream = DEFAULT_LoggingStream;
	pDefaultConnectionParameters->sClientHeader = NULL;
	pDefaultConnectionParameters->sProtocol = DEFAULT_Protocol;
	pDefaultConnectionParameters->sUserAgent = sDefaultUserAgent;
	pDefaultConnectionParameters->cook = NULL;
	pDefaultConnectionParameters->sVerify = 0;
	pDefaultConnectionParameters->sCompress = NULL;
	/* gConversionFormats->ctFormats                      = 0; -- not used */
	*hDefaultConnectionParameters = pDefaultConnectionParameters;

	/* libcurl initialization must be done exactly once during life of program */
	curl_global_init(CURL_GLOBAL_ALL);

	/* Add a new parameter to DefaultConnectionParameters to hold a libcurl
	 * handle. Create the handle here and use elsewhere.
	 */

	return SUCCESS;
}

int parse_buffer(void *pBuffer) {
	int zScannerLocked;
	int zParseResult;
	zScannerLocked = pthread_mutex_lock(&mutexInitializer); /* this call will block if another thread is already using the scanner */
	vResetScanner();
	zParseResult = ddsparse(pBuffer);
	zScannerLocked = pthread_mutex_unlock(&mutexInitializer);
	return zParseResult;
}

size_t curl_callback_data(void *curl_buffer, size_t size, size_t nmemb,
		void *pParserBuffer) {
	ParserBuffer *parser_buffer;
	int lenCurlBuffer = size * nmemb;
	Buffer *byte_buffer;

	parser_buffer = (ParserBuffer *) pParserBuffer;
	byte_buffer = parser_buffer->pInputBuffer;

	if (lenCurlBuffer <= 0) {
		parser_buffer->sError = "data callback occurred with no data";
		return 0;
	}

	if (bufferAppend(byte_buffer, curl_buffer, 0, lenCurlBuffer)
			!= lenCurlBuffer) {
		parser_buffer->sError = "out of memory";
		return 0;
	}

	DBG(printf("appended %i chars to buffer:\n", lenCurlBuffer));

	/* report progress if requested */
#ifdef gpvUpdateProgress
	if (gpvUpdateProgress != NULL)
	gpvUpdateProgress(byte_buffer->iContentLength, 0, "received");
#endif

	/* handle data */
	if (parser_buffer->eReadMode == RM_Buffered) {
		int posDataHeader;
		DBG(printf("mode is buffered, content position is %i, length is %i\n",
				byte_buffer->iContentPosition0, byte_buffer->iContentLength));
		posDataHeader = indexOf_binary(byte_buffer->pcharContent
				+ byte_buffer->iContentPosition0, byte_buffer->iContentLength,
		DATA_HEADER, strlen(DATA_HEADER));
		if (posDataHeader > 0) { /* then this is a DataDDS; switch reading mode and start reading data */
			int parser_return_value;
			int posData = posDataHeader + strlen(DATA_HEADER);
			int posDataLength = byte_buffer->iContentLength - posData + 1; /* remember how much data there is so far */

			DBG(printf("data header found\n"));

			byte_buffer->iContentLength = posDataHeader
					- byte_buffer->iContentPosition0; /* truncate to just the DDS and include the terminal newline */

			/* read the DDS portion of the data */
			parser_return_value = parse_buffer(parser_buffer);
			if (parser_return_value == 0) { /* bison parser returns 0 on success, 1 on failure */
				if (parser_buffer->pOStructure == NULL) {
					parser_buffer->sError
							= "Internal error, parser returned true but structure does not exist";
					return 0;
				}
			} else {
				return 0; /* error will be in the parser buffer */
			}

			/* set the structure type to DataDDS and change mode to data reading */
			parser_buffer->pOStructure->eStructureType = STRUCTURE_DataDDS;
			parser_buffer->eReadMode = RM_Data;
			byte_buffer->iContentPosition0 = posData - 1; /* posData is one-based, iContentPosition0 is zero based */
			byte_buffer->iContentLength = posDataLength;

			return lenCurlBuffer;
		} else {
			DBG(printf("no data header\n"));
		}
		return lenCurlBuffer;
	} else if (parser_buffer->eReadMode == RM_Data) {
		DBG(printf("continuing data read\n"));

		return lenCurlBuffer;
	} else {
		parser_buffer->sError = "unknown read mode";
		return 0;
	}
}

/* loads the returned raw data/text from the URL into a buffer */
eRESULT zConnection_Fetch(char *sURL, Buffer * pBuffer,
			  ThreadSpecificStorage * pTSS, StringBuffer * sbError) {
    URL url;
    ConnectionParameters *parameters;
    ConnectionParameters *pDefaultConnectionParameters;

    char *pURL;
    char **sUser;
    char **sPassword;
    char **sError;

    if (sURL == NULL) {
	sbAppend(sbError, "internal error, no URL string provided");
	return FAILURE;
    }
    if (pBuffer == NULL) {
	sbAppend(sbError, "internal error, no buffer");
	return FAILURE;
    }
    if (pTSS == NULL) {
	sbAppend(sbError, "internal error, no thread-specific storage");
	return FAILURE;
    }
    pDefaultConnectionParameters = pTSS->pConnectionParameters;
    if (pDefaultConnectionParameters == NULL) {
	sbAppend(sbError, "internal error, no default connection parameters");
	return FAILURE;
    }
    if(strchr(sURL, '@')){
	/* sURL --> &sURL to match formal parameter. jhrg 11/20/08 */
	if (parseURL(pURL, sUser, &sURL, sPassword,
		     sError)== FAILURE){
	    return FAILURE;
	}

    }
    /* initialize the URL */
    url.sRawURL = sURL;
    url.eRequestClass = RC_Raw;
    url.sCanonicalURL = sURL;

    /* initialize the parameters -- todo make init method */
    parameters = malloc(sizeof(ConnectionParameters));
    if (parameters == NULL) {
	sbAppend(sbError, "out of memory trying to allocate parameters");
	return FAILURE;
    }
    parameters->pstructProxy = pDefaultConnectionParameters->pstructProxy;
    if(pDefaultConnectionParameters->pstructProxy != NULL) {
	parameters->pstructProxy->sHost = pDefaultConnectionParameters->pstructProxy->sHost;
	parameters->pstructProxy->iPort = pDefaultConnectionParameters->pstructProxy->iPort;
	parameters->pstructProxy->sUser = pDefaultConnectionParameters->pstructProxy->sUser;
	parameters->pstructProxy->sPassword = pDefaultConnectionParameters->pstructProxy->sPassword;
    }
    parameters->pstructURL = &url;
    parameters->sUserC = pDefaultConnectionParameters->sUserC;
    parameters->sPasswordC = pDefaultConnectionParameters->sPasswordC;
    parameters->iConnectTimeout_seconds
	= pDefaultConnectionParameters->iConnectTimeout_seconds;
    parameters->iReadTimeout_seconds
	= pDefaultConnectionParameters->iReadTimeout_seconds;
    parameters->sProtocol = pDefaultConnectionParameters->sProtocol;
    parameters->sUserAgent = pDefaultConnectionParameters->sUserAgent;
    parameters->sClientHeader = pDefaultConnectionParameters->sClientHeader;
    parameters->pLoggingStream = pDefaultConnectionParameters->pLoggingStream;
    parameters->cook = pDefaultConnectionParameters->cook;
    parameters->sVerify = pDefaultConnectionParameters->sVerify;
    parameters->sCompress = pDefaultConnectionParameters->sCompress;

    return zConnection_FetchWithParameters(parameters, pBuffer, sbError);
}

eRESULT zConnection_FetchWithParameters(ConnectionParameters * parameters,
		Buffer * pBuffer, StringBuffer * sbError) {
	eRESULT result;
	ParserBuffer *pParserBuffer = malloc(sizeof(ParserBuffer));

	if (pParserBuffer == NULL) {
		sbAppend(sbError, "unable to create parser buffer, out of memory");
		return FAILURE;
	}
	pParserBuffer->iStatus = TRUE;
#if 0
	pParserBuffer->eGridMode = GM_undefined;
#endif
	pParserBuffer->eReadMode = RM_Buffered;
	pParserBuffer->pOStructure = NULL;
	pParserBuffer->sError = NULL;
	pParserBuffer->pInputBuffer = pBuffer;
	pParserBuffer->eStructureType = STRUCTURE_undefined;
	pParserBuffer->pCurlRequest = NULL;

	if (zFetch_Synchronous(parameters, pParserBuffer, sbError)) {
		result = SUCCESS;
	} else {
		result = FAILURE;
		sbInsert(sbError, "failed to fetch: ");
	}
	free(pParserBuffer); /* note only container is freed, buffer inside container continues to live */
	return result;
}

/* opens a connection to the URL and loads an OPeNDAP object, returns an error
 * if the connection does not return a valid OPeNDAP stream or if there is a server errror */
eRESULT zConnection_Load(char *sURL, DAPObject ** hStructure,
		ThreadSpecificStorage * pTSS, StringBuffer * sbError) {

	URL url;
	ConnectionParameters *parameters;
	ConnectionParameters *pDefaultConnectionParameters;
	Buffer *pInputBuffer;
	ParserBuffer *pParserBuffer;
	char *sServerError;
	eRESULT result;
	DAPNode *nodeRoot;
	char *sCurrentBufferContent;

	if (sURL == NULL) {
		sbAppend(sbError, "internal error, no URL string provided");
		return FAILURE;
	}
	if (hStructure == NULL) {
		sbAppend(sbError, "internal error, no structure handle provided");
		return FAILURE;
	}
	if (pTSS == NULL) {
		sbAppend(sbError, "internal error, no thread-specific storage");
		return FAILURE;
	}
	pDefaultConnectionParameters = pTSS->pConnectionParameters;
	if (pDefaultConnectionParameters == NULL) {
		sbAppend(sbError, "internal error, no default connection parameters");
		return FAILURE;
	}

	/* initialize the URL */
	url.sRawURL = sURL;
	url.eRequestClass = RC_Raw;
	url.sCanonicalURL = sURL;

	/* initialize the parameters -- todo make init method */
	parameters = malloc(sizeof(ConnectionParameters));
	if (parameters == NULL) {
		sbAppend(sbError, "out of memory trying to allocate parameters");
		return FAILURE;
	}
	parameters->pstructProxy = pDefaultConnectionParameters->pstructProxy;
	if(pDefaultConnectionParameters->pstructProxy != NULL) {
		parameters->pstructProxy->sHost = pDefaultConnectionParameters->pstructProxy->sHost;
	    parameters->pstructProxy->iPort = pDefaultConnectionParameters->pstructProxy->iPort;
	    parameters->pstructProxy->sUser = pDefaultConnectionParameters->pstructProxy->sUser;
	    parameters->pstructProxy->sPassword = pDefaultConnectionParameters->pstructProxy->sPassword;
	}
	parameters->pstructURL = &url;
	parameters->sUserC = pDefaultConnectionParameters->sUserC;
	parameters->sPasswordC = pDefaultConnectionParameters->sPasswordC;
	parameters->iConnectTimeout_seconds
			= pDefaultConnectionParameters->iConnectTimeout_seconds;
	parameters->iReadTimeout_seconds
			= pDefaultConnectionParameters->iReadTimeout_seconds;
	parameters->sProtocol = pDefaultConnectionParameters->sProtocol;
	parameters->sUserAgent = pDefaultConnectionParameters->sUserAgent;
	parameters->sClientHeader = pDefaultConnectionParameters->sClientHeader;
	parameters->pLoggingStream = pDefaultConnectionParameters->pLoggingStream;
	parameters->cook = pDefaultConnectionParameters->cook;
	parameters->sVerify = pDefaultConnectionParameters->sVerify;
	parameters->sCompress = pDefaultConnectionParameters->sCompress;

	pInputBuffer = bufferNew(0);DBG(printf("allocated input buffer: %p", pInputBuffer));
	if (pInputBuffer == NULL) {
		sbAppend(sbError,
				"failed to create new input buffer, probably out of memory");
		free(parameters);DBG(printf("freeing input buffer1: %p", pInputBuffer));
		imp_buffer_Free(pInputBuffer /* not used jhrg 4/3/07, sbError */); /* ignore any errors -- todo generate warning */
		return FAILURE;
	}

	pParserBuffer = malloc(sizeof(ParserBuffer));
	if (pParserBuffer == NULL) {
		sbAppend(sbError, "unable to create parser buffer, out of memory?");
		free(parameters);
        DBG(printf("freeing input buffer2: %p", pInputBuffer);
		)imp_buffer_Free(pInputBuffer /* not used jhrg 4/3/07, sbError */); /* ignore any errors -- todo generate warning */
		return FAILURE;
	}
	pParserBuffer->iStatus = TRUE;
#if 0
	pParserBuffer->eGridMode = GM_undefined;
#endif
	pParserBuffer->eReadMode = RM_Buffered;
	pParserBuffer->pOStructure = NULL;
	pParserBuffer->sError = NULL;
	pParserBuffer->pInputBuffer = pInputBuffer;
	pParserBuffer->eStructureType = STRUCTURE_undefined;
	pParserBuffer->pCurlRequest = NULL;

	if (zFetch_Synchronous(parameters, pParserBuffer, sbError) == FAILURE) {
		sbInsert(sbError, "failed to fetch: ");
		goto FAIL;
	}

	/*
	 sCurrentBufferContent = pInputBuffer->pcharContent + pInputBuffer->iContentPosition0;
	 printf("buffer length is %i\n", pInputBuffer->iContentLength);
	 printf("buffer content is %s\n", sCurrentBufferContent );
	 */

	/* Based on a warning from gcc, I replaced the calls with
	 sServerError to ones with NULL. The variable is not
	 set/initialized until later in the function and sbApend_string()
	 tests for a null string and substitutes the vale "[null]". jhrg
	 4/2/07 */

	if (pParserBuffer->pInputBuffer == NULL) {
		sbAppend_string(sbError, "Input buffer unexpectedly does not exist",
		/* sServerError */NULL);
		goto FAIL;
	}
	if (pParserBuffer->pInputBuffer->pcharContent == NULL) {
		sbAppend_string(sbError, "Server response content is null",
		/* sServerError */NULL);
		goto FAIL;
	}

	sCurrentBufferContent = pParserBuffer->pInputBuffer->pcharContent
			+ pParserBuffer->pInputBuffer->iContentPosition0;
	sServerError = findError(sCurrentBufferContent);
	if (sServerError != NULL) {
		sbAppend_string(sbError, "Server returned error: %s", sServerError);
		goto FAIL;
	}

	/* parse the content, if the data read function did not already do it */
	if (pParserBuffer->pOStructure == NULL) { /* then it is a DDS or DAS */

		/* determine whether it is a DDS or DAS
		 * This clause should be eliminated when code has been added to the callback
		 * routine to determine the structure type by header description
		 */
		if (bufferStartsWith(pInputBuffer, "DATASET")) {
			pParserBuffer->eStructureType = STRUCTURE_DDS;
		} else if (bufferStartsWith(pInputBuffer, "ATTRIBUTES")) {
			pParserBuffer->eStructureType = STRUCTURE_DAS;
		}

		if (pParserBuffer->eStructureType == STRUCTURE_DDS) {
			int parser_return_value;
			parser_return_value = parse_buffer((void *) pParserBuffer);
			if (DEBUG)
				printf("parsing DDS buffer\n");
			if (parser_return_value == 0) { /* bison parser returns 0 on success, 1 on failure */
				if (pParserBuffer->pOStructure == NULL) {
					sbInsert_string(
							sbError,
							"Internal error, parser returned true but structure does not exist (%s)",
							sURL);
					goto FAIL;
				} else {
					/* At this point eStructureType has already been set. jhrg 12/02/08 */
					pParserBuffer->pOStructure->eStructureType = STRUCTURE_DDS; /* todo support other object types */
					*hStructure = pParserBuffer->pOStructure;
					goto SUCCEED;
				}
			} else {
				sbInsert_string(sbError, "Error parsing: %s",
						pParserBuffer->sError);
				goto FAIL;
			}
		} else if (pParserBuffer->eStructureType == STRUCTURE_DAS) {
			if (zParseDAS(hStructure, pParserBuffer->pInputBuffer, sbError)
					== SUCCESS) {
				goto SUCCEED;
			} else {
				sbInsert(sbError, "Error parsing: ");
				goto FAIL;
			}
		} else {
			sbAppend(sbError, "unable to determine structure type");
			goto FAIL;
		}
	}

	/* Read the data -- this is a temporary (hopefully) clause that reads the data
	 * if necessary. In the future this will be managed by the callback routine
	 * curl_callback_data above which will call ReadData to dynamically read data
	 * as it comes in. Right now the entire buffer is read in and now the data read
	 * is done all at once at the end, ie right here.
	 */
	nodeRoot = pParserBuffer->pOStructure->nodeRoot;
	if (nodeRoot == NULL) {
		sbAppend(sbError, "about to read data but root node unexpectedly null");
		goto FAIL;
	}
	/*
	 printf("buffer content %p:\n", pInputBuffer);
	 bufferDump( pInputBuffer, stdout, 1, 500 );
	 */

	if (read_Node(nodeRoot, NULL, pInputBuffer, sbError) == SUCCESS) {
		pParserBuffer->pOStructure->eStructureType = STRUCTURE_DataDDS;
		*hStructure = pParserBuffer->pOStructure;
		goto SUCCEED;
	} else {
		sbInsert(sbError, "error reading data: ");
		goto FAIL;
	}

	FAIL: result = FAILURE;
	goto FINALLY;
	SUCCEED: result = SUCCESS;
	goto FINALLY;
	FINALLY: free(parameters);DBG	(printf("freeing input buffer3: %p", pInputBuffer);
	)
	if (imp_buffer_Free(pInputBuffer /* not used jhrg 4/3/07, sbError */) ==
			FAILURE) {fputs("error freeing data buffer", stderr);
	}
	if (pParserBuffer != NULL)
	free(pParserBuffer);
	return result;
}

eRESULT zFetch_Synchronous(ConnectionParameters * parameters,
		ParserBuffer * pParserBuffer, StringBuffer * sbError) {

	URL *url;
	CURL *pCurlRequest;
	char *sURL;
	char *sUserAgent;
	CURLcode curl_result;

	/* validate parameters */
	if (parameters == NULL) {
		sbAppend(sbError, "internal error, no connection parameters supplied");
		return FAILURE;
	}
	url = parameters->pstructURL;
	if (url == NULL) {
		sbAppend(sbError, "no URL");
		return FAILURE;
	}
	if (url->sCanonicalURL == NULL) {
		sbAppend(sbError, "URL bad/unresolved");
		return FAILURE;
	}
	sUserAgent = parameters->sUserAgent;
	if (sUserAgent == NULL) {
		sbAppend(sbError, "internal error, no user agent supplied");
		return FAILURE;
	}

	if (pParserBuffer == NULL) {
		sbAppend(sbError, "internal error, no buffer");
		return FAILURE;
	}

	/* set up a URL request */
	pCurlRequest = curl_easy_init();
	if (!pCurlRequest) {
		sbAppend(sbError, "unable to initialize transfer request");
		return FAILURE;
	}
	sURL = url->sCanonicalURL;
	/* curl_easy_setopt(pCurlRequest, CURLOPT_VERBOSE, TRUE); */
	curl_easy_setopt(pCurlRequest, CURLOPT_URL, sURL);

	/*Added this to be able to turn ssl configuration off*/

	if(parameters->sVerify){
	curl_easy_setopt(pCurlRequest, CURLOPT_SSL_VERIFYPEER, 0);
	curl_easy_setopt(pCurlRequest, CURLOPT_SSL_VERIFYHOST, 0);
	}
	if(parameters->sCompress){
#ifdef CURLOPT_ENCODING
	curl_easy_setopt(pCurlRequest, CURLOPT_ENCODING, 'deflate, gzip');
#endif
	}
	curl_easy_setopt(pCurlRequest, CURLOPT_USERAGENT, sUserAgent);
	curl_easy_setopt(pCurlRequest, CURLOPT_WRITEFUNCTION, curl_callback_data);
	curl_easy_setopt(pCurlRequest, CURLOPT_WRITEDATA, pParserBuffer);

	if (parameters->pstructProxy) {
#if 0
		fprintf(stderr, "Setting proxy host and port: %s:%d\n",
				parameters->pstructProxy->sHost, parameters->pstructProxy->iPort);
#endif
		curl_easy_setopt(pCurlRequest, CURLOPT_PROXY, parameters->pstructProxy->sHost);
		curl_easy_setopt(pCurlRequest, CURLOPT_PROXYPORT, parameters->pstructProxy->iPort);
		if (parameters->pstructProxy->sUser) {
			int iUserPassSize = strlen(parameters->pstructProxy->sUser)
				+ strlen(parameters->pstructProxy->sPassword) +2;
			char *sUserPassword = malloc(sizeof(char) * iUserPassSize);
			if (!sUserPassword) {
				sbAppend(sbError, "Failed to allocate memory for the proxy username and/or password.");
				return FAILURE;
			}
			strncpy(sUserPassword, parameters->pstructProxy->sUser, iUserPassSize);
			strncat(sUserPassword, ":", iUserPassSize);
			strncat(sUserPassword, parameters->pstructProxy->sPassword, iUserPassSize);
			curl_easy_setopt(pCurlRequest, CURLOPT_PROXYUSERPWD, sUserPassword);
#ifdef CURLOPT_PROXYAUTH
	        curl_easy_setopt(pCurlRequest, CURLOPT_PROXYAUTH, (long)CURLAUTH_ANY);
#endif



			free(sUserPassword);
		}
	}
	if(parameters->sUserC){
		int iUserPassSize = strlen(parameters->sUserC)
			+ strlen(parameters->sPasswordC) +2;
		char *sUserPassword = malloc(sizeof(char) * iUserPassSize);
		if (!sUserPassword) {
			sbAppend(sbError, "Failed to allocate memory for the username and/or password.");
			return FAILURE;
		}
		strncpy(sUserPassword, parameters->sUserC, iUserPassSize);
		strncat(sUserPassword, ":", iUserPassSize);
		strncat(sUserPassword, parameters->sPasswordC, iUserPassSize);
		curl_easy_setopt(pCurlRequest, CURLOPT_USERPWD, sUserPassword);
		curl_easy_setopt(pCurlRequest, CURLOPT_HTTPAUTH,(long)CURLAUTH_ANY);

		free(sUserPassword);
	}
	if(parameters->cook){
		curl_easy_setopt(pCurlRequest, CURLOPT_COOKIEJAR, parameters->cook);
		curl_easy_setopt(pCurlRequest, CURLOPT_COOKIESESSION, 1);
	}

	fflush(stdout);
	fflush(stderr);
	pParserBuffer->pCurlRequest = pCurlRequest;
	curl_result = curl_easy_perform(pCurlRequest);
	curl_easy_cleanup(pCurlRequest);
	if (curl_result != CURLE_OK) {
		if (curl_result == CURLE_WRITE_ERROR) { /* a parsing/handling error has occurred */
			sbAppend_string(sbError, "parsing error: %s", pParserBuffer->sError);
		} else {
			sbAppend_int(sbError, "cURL error code %d: ", curl_result);
			sbAppend(sbError, sCURLError(curl_result));
		}
		return FAILURE;
	}

	return SUCCESS;
}
#if 0
/* currently not used, but may be used in the future */
eRESULT zFetch_Asynchronous(ConnectionParameters * parameters,
		ParserBuffer * pParserBuffer, StringBuffer * sbError) {
	URL *url;
	CURL *pCurlRequest;
	CURLM *pMultiEnvironment;
	int ctRunningRequests;
	char *sURL;
	char *sUserAgent;

	if (parameters == NULL) {
		sbAppend(sbError, "internal error, no connection parameters supplied");
		return FAILURE;
	}
	url = parameters->pstructURL;
	if (url == NULL) {
		sbAppend(sbError, "no URL");
		return FAILURE;
	}
	if (url->sCanonicalURL == NULL) {
		sbAppend(sbError, "URL bad/unresolved");
		return FAILURE;
	}
	sUserAgent = parameters->sUserAgent;
	if (sUserAgent == NULL) {
		sbAppend(sbError, "internal error, no user agent supplied");
		return FAILURE;
	}

	pCurlRequest = curl_easy_init();
	if (pCurlRequest == NULL) {
		sbAppend(sbError, "failed to initialize request object");
		return FAILURE;
	}

	if (pParserBuffer == NULL) {
		sbAppend(sbError, "internal error, no buffer");
		return FAILURE;
	}

	/* set up a URL request */
	sURL = url->sCanonicalURL;
	curl_easy_setopt(pCurlRequest, CURLOPT_URL, sURL);
	curl_easy_setopt(pCurlRequest, CURLOPT_USERAGENT, sUserAgent);
	curl_easy_setopt(pCurlRequest, CURLOPT_WRITEFUNCTION, curl_callback_data);
	curl_easy_setopt(pCurlRequest, CURLOPT_WRITEDATA, pParserBuffer); /* CURL will write to connection buffer */

	pMultiEnvironment = curl_multi_init(); /* init a multi stack */
	pParserBuffer->pCurlRequest = pCurlRequest;
	curl_multi_add_handle(pMultiEnvironment, pCurlRequest);

	while (CURLM_CALL_MULTI_PERFORM == curl_multi_perform(pMultiEnvironment,
			&ctRunningRequests)) {

		/* while CURL is performing callbacks will be made to curl_callback_data */
		printf("performing request, received: %d\n",
				pParserBuffer->pInputBuffer->iContentLength);
	}

	{ /* See how the transfer(s) went */
		CURLMsg *pTransferMessage;
		int ctMessagesRemaining;
		while ((pTransferMessage = curl_multi_info_read(pMultiEnvironment,
				&ctMessagesRemaining))) {
			if (pTransferMessage->msg == CURLMSG_DONE) {
				if (pTransferMessage->easy_handle == pCurlRequest) {
					printf("HTTP transfer completed with status %d\n",
							pTransferMessage->data.result);
					break;
				}
			}
		}
	}

	curl_multi_cleanup(pMultiEnvironment);

	curl_easy_cleanup(pCurlRequest);

	return SUCCESS;

}
#endif
char *sCURLError(int iErrorCode) {
	switch (iErrorCode) {
	case CURLE_OK:
		return "OK";
	case CURLE_UNSUPPORTED_PROTOCOL:
		return "Unsupported Protocol. The URL you passed to libcurl used a protocol that this libcurl does not support. The support might be a compile-time option that you didn't use, it can be a misspelled protocol string or just a protocol libcurl has no code for.";
	case CURLE_FAILED_INIT:
		return "Very early initialization code failed. This is likely to be an internal error or problem.";
	case CURLE_URL_MALFORMAT:
		return "The URL was not properly formatted.";
	case CURLE_URL_MALFORMAT_USER:
		return "URL user malformatted. The user-part of the URL syntax was not correct.";
	case CURLE_COULDNT_RESOLVE_PROXY:
		return "Couldn't resolve proxy. The given proxy host could not be resolved.";
	case CURLE_COULDNT_RESOLVE_HOST:
		return "Couldn't resolve host. The given remote host was not resolved.";
	case CURLE_COULDNT_CONNECT:
		return "Failed to connect() to host or proxy.";
	case CURLE_FTP_WEIRD_SERVER_REPLY:
		return "After connecting to an FTP server, libcurl expects to get a certain reply back. This error code implies that it god a strange or bad reply. The given remote server is probably not an OK FTP server.";
	case CURLE_FTP_ACCESS_DENIED:
		return "Denied access when trying to login to an FTP server or when trying to change working directory to the one given in the URL.";
	case CURLE_FTP_USER_PASSWORD_INCORRECT:
		return "The FTP server rejected access to the server after the password was sent to it. It might be because the username or the password were incorrect or just that the server is not allowing you access for the moment etc.";
	case CURLE_FTP_WEIRD_PASS_REPLY:
		return "After having sent the FTP password to the server, libcurl expects a proper reply. This error code indicates that an unexpected code was returned.";
	case CURLE_FTP_WEIRD_USER_REPLY:
		return "After having sent user name to the FTP server, libcurl expects a proper reply. This error code indicates that an unexpected code was returned.";
	case CURLE_FTP_WEIRD_PASV_REPLY:
		return "libcurl failed to get a sensible result back from the server as a response to either a PASV or a EPSV command. The server is flawed.";
	case CURLE_FTP_WEIRD_227_FORMAT:
		return "FTP servers return a 227-line as a response to a PASV command. If libcurl fails to parse that line, this return code is passed back.";
	case CURLE_FTP_CANT_GET_HOST:
		return "An internal failure to lookup the host used for the new connection.";
	case CURLE_FTP_CANT_RECONNECT:
		return "A bad return code on either PASV or EPSV was sent by the FTP server, preventing libcurl from being able to continue.";
	case CURLE_FTP_COULDNT_SET_BINARY:
		return "Received an error when trying to set the transfer mode to binary.";
	case CURLE_PARTIAL_FILE:
		return "A file transfer was shorter or larger than expected. This happens when the server first reports an expected transfer size, and then delivers data that doesn't match the previously given size.";
	case CURLE_FTP_COULDNT_RETR_FILE:
		return "This was either a weird reply to a 'RETR' command or a zero byte transfer complete.";
	case CURLE_FTP_WRITE_ERROR:
		return "After a completed file transfer, the FTP server did not respond a proper 'transfer successful' code.";
	case CURLE_FTP_QUOTE_ERROR:
		return "When sending custom 'QUOTE' commands to the remote server, one of the commands returned an error code that was 400 or higher.";
	case CURLE_HTTP_RETURNED_ERROR:
		return "This is returned if CURLOPT_FAILONERROR is set TRUE and the HTTP server returns an error code that is >= 400.";
	case CURLE_WRITE_ERROR:
		return "An error occurred when writing received data to a local file, or an error was returned to libcurl from a write callback.";
	case CURLE_MALFORMAT_USER:
		return "Malformat user. User name badly specified. *Not currently used*";
	case CURLE_FTP_COULDNT_STOR_FILE:
		return "FTP couldn't STOR file. The server denied the STOR operation. The error buffer usually contains the server's explanation to this.";
	case CURLE_READ_ERROR:
		return "There was a problem reading a local file or an error returned by the read callback.";
	case CURLE_OUT_OF_MEMORY:
		return "Out of memory. A memory allocation request failed. May result in system instability.";
	case CURLE_OPERATION_TIMEOUTED:
		return "Operation timeout. The specified timeout period was reached according to the conditions.";
	case CURLE_FTP_COULDNT_SET_ASCII:
		return "libcurl failed to set ASCII transfer type (TYPE A).";
	case CURLE_FTP_PORT_FAILED:
		return "The FTP PORT command returned error. This mostly happen when you haven't specified a good enough address for libcurl to use. See CURLOPT_FTPPORT.";
	case CURLE_FTP_COULDNT_USE_REST:
		return "The FTP REST command returned error. This should never happen unless the FTP server has an internal problem or bad implementation.";
	case CURLE_FTP_COULDNT_GET_SIZE:
		return "The FTP SIZE command returned error. The SIZE command may not be allowed by this server.";
	case CURLE_HTTP_RANGE_ERROR:
		return "The HTTP server does not support or accept range requests.";
	case CURLE_HTTP_POST_ERROR:
		return "HTTP post error due to internal problems.";
	case CURLE_SSL_CONNECT_ERROR:
		return "SSL Connect Error; a problem occurred somewhere in the SSL/TLS handshake. You really want the error buffer and read the message there as it pinpoints the problem slightly more. Could be certificates (file formats, paths, permissions), passwords, and others.";
	case CURLE_FTP_BAD_DOWNLOAD_RESUME:
		return "Bad Download Resume; attempted FTP resume beyond file size.";
	case CURLE_FILE_COULDNT_READ_FILE:
		return "Could not read file; a file given with FILE:// couldn't be opened. Most likely because the file path doesn't identify an existing file or file permissions prevent it from being accessed.";
	case CURLE_LDAP_CANNOT_BIND:
		return "LDAP cannot bind. LDAP bind operation failed.";
	case CURLE_LDAP_SEARCH_FAILED:
		return "LDAP search failed.";
	case CURLE_LIBRARY_NOT_FOUND:
		return "LDAP library not found.";
	case CURLE_FUNCTION_NOT_FOUND:
		return "Required LDAP function not found.";
	case CURLE_ABORTED_BY_CALLBACK:
		return "Aborted by callback. A callback returned 'abort' to libcurl.";
	case CURLE_BAD_FUNCTION_ARGUMENT:
		return "Internal error. A function was called with a bad parameter.";
	case CURLE_BAD_CALLING_ORDER:
		return "Internal error. A function was called in a bad order.";
	case CURLE_HTTP_PORT_FAILED:
		return "Interface error. A specified outgoing interface could not be used. Set which interface to use for outgoing connections' source IP address with CURLOPT_INTERFACE.";
	case CURLE_BAD_PASSWORD_ENTERED:
		return "Bad password entered. An error was signaled when the password was entered. This can also be the result of a 'bad password' returned from a specified password callback.";
	case CURLE_TOO_MANY_REDIRECTS:
		return "Too many redirects. When following redirects, libcurl hit the maximum amount. Set your limit with CURLOPT_MAXREDIRS.";
	case CURLE_UNKNOWN_TELNET_OPTION:
		return "An option set with CURLOPT_TELNETOPTIONS was not recognized/known. Refer to the appropriate documentation.";
	case CURLE_TELNET_OPTION_SYNTAX:
		return "A telnet option string was illegally formatted.";
	case CURLE_OBSOLETE:
		return "This is not an error. This used to be another error code in an old libcurl version and is currently unused.";
	case CURLE_SSL_PEER_CERTIFICATE:
		return "The remote server's SSL certificate was deemed not OK.";
	case CURLE_GOT_NOTHING:
		return "Nothing was returned from the server, and under the circumstances, getting nothing is considered an error.";
	case CURLE_SSL_ENGINE_NOTFOUND:
		return "The specified crypto engine wasn't found.";
	case CURLE_SSL_ENGINE_SETFAILED:
		return "Failed setting the selected SSL crypto engine as default!";
	case CURLE_SEND_ERROR:
		return "Failed sending network data.";
	case CURLE_RECV_ERROR:
		return "Failure receiving network data.";
	case CURLE_SHARE_IN_USE:
		return "Share is in use.";
	case CURLE_SSL_CERTPROBLEM:
		return "problem with the local client certificate";
	case CURLE_SSL_CIPHER:
		return "specified cipher unusable";
	case CURLE_SSL_CACERT:
		return "problem with the CA cert (path? access rights?)";
	case CURLE_BAD_CONTENT_ENCODING:
		return "Unrecognized transfer encoding";

		/* The following errors are not present in some cURL versions.
		 * Here they are omitted because they are not used for the cURL functionality
		 * used by the OCAPI anyway.
		 case CURLE_LDAP_INVALID_URL: return "Invalid LDAP URL";
		 case CURLE_FILESIZE_EXCEEDED: return "Maximum file size exceeded";
		 case CURLE_FTP_SSL_FAILED: return "Requested FTP SSL level failed";
		 */
	default:
		return "[no description for error code]";
	}
}
