/*
 * This file is part of the "Ocapi" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: Dennis Heimbigner <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

#ifndef PARSE_H
#define PARSE_H

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "global.h"
#include "stack.h"

extern StringBuffer* sbError;
extern ParserBuffer* dapParserState;
extern int dap_datahdr;

extern Stack* dapdimnames;
extern Stack* dapdimsizes;
extern Stack* daptemp;
extern Stack* dapvalues;
extern Stack* dapstack;

/* Primary entry points from rest of OCAPI
* For DDS parsing */
extern int ddsparse(void* parserstate); /* dds parser wrap*/
/* For DAS parsing */
eRESULT zParseDAS(DAPObject** rootp , Buffer *pBuffer, StringBuffer *sberr );


/* dap.y*/
extern void dap_parse_error(char* msg, char* token, int lineno);
extern void dap_cleanup(void);
extern int daperror(char*);
extern int dapparse(void);

extern int OCAPI_Input( char* buf, unsigned int max_size );
/*extern void reset(void);*/
extern void vResetScanner(void);
extern void dap_fatal_error(char*);
int check_int32(char*);

/* dap.lex*/
extern int dap_line_num;
extern int daplex(void);

/* build.c*/
extern DAPNode* newDAPNode(eNODE_TYPE t, Stack* subnodes, char* name);
extern eARRAY_TYPE arrayTypeOf(eNODE_TYPE t);
extern NodeContent_Array* makeNCArray(eNODE_TYPE t, char* name);
extern NodeContent_Array* cloneNCArray(NodeContent_Array* nca);
extern void makeArray(char* name, eNODE_TYPE basetype);
extern DAPNode* makeArray0(char* name0, eNODE_TYPE basetype, int);
extern void makeStructure(char* name0, int count);
extern void makeGrid(char* name0, int count);
extern void makeSequence(char* name0, int count);
extern void makeRoot(eSTRUCTURE_TYPE, char* name0, int count);
extern void add_attributevalue(char* value, eNODE_TYPE etype);
extern int checktype(char* value, eNODE_TYPE etype);
extern void makeAttr(char* name0, eATTRIBUTE_TYPE etype);
extern void makeAttrnode(char* name0, int count);
extern void** makeValuelist(eATTRIBUTE_TYPE);
void makeAttrroot(int stackbase);
int checkdupattr(int stackbase);

#define ID_MAX 1024
#define STACKALLOC 1024
#define ATTRTAG  0x0A770000
#define ATTRMASK 0xFFFF0000

#define nulldup(s) (s==NULL?NULL:strdup(s))
#define nullstring(s) (s==NULL?"null":s)

#endif /*PARSE_H*/
