/*
 * This file is part of the "Ocapi" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: Dennis Heimbigner <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

#ifndef STACK_H
#define STACK_H

// Define the type of the elements in the stack

typedef unsigned long elem_t; // the type of the contents array; must be at least
                      // size of a pointer.
#define ELEMNULL ((elem_t)0)

typedef struct Stack {
  long alloc;
  long length;
  elem_t* content;
} Stack;

#ifndef EXTERNC
#if defined(_CPLUSPLUS_) || defined(__CPLUSPLUS__) || defined(__cplusplus)
#define EXTERNC extern "C"
#else
#define EXTERNC extern
#endif
#endif

EXTERNC Stack* stackNew(long initial);
EXTERNC int stackFree(Stack*);
EXTERNC int stackClear(Stack*);
EXTERNC int stackDepth(Stack*);
EXTERNC int stackSetdepth(Stack*,long depth);

/* Return the ith element of stack: 0 is bottom element*/
EXTERNC elem_t stackGet(Stack*,long);

EXTERNC int stackPush(Stack*,elem_t); /* Add at Tail */
EXTERNC elem_t stackPop(Stack*);
EXTERNC elem_t stackTop(Stack*);

int zTestStack();

#endif /*STACK_H*/
